@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2013-2016 Joris Vink <joris@coders.se>
+ * Copyright (c) 2013-2018 Joris Vink <joris@coders.se>
  *
  * Permission to use, copy, modify, and distribute this software for any
  * purpose with or without fee is hereby granted, provided that the above
@@ -53,19 +53,22 @@ extern "C" {
 extern int daemon(int, int);
 #endif
 
+#if !defined(KORE_NO_SENDFILE) && defined(KORE_NO_TLS)
+#if defined(__MACH__) || defined(__FreeBSD_version) || defined(__linux__)
+#define KORE_USE_PLATFORM_SENDFILE	1
+#endif
+#endif
+
 #define KORE_RESULT_ERROR	0
 #define KORE_RESULT_OK		1
 #define KORE_RESULT_RETRY	2
 
-#define KORE_VERSION_MAJOR	2
-#define KORE_VERSION_MINOR	0
-#define KORE_VERSION_PATCH	0
-#define KORE_VERSION_STATE	"release"
-
 #define KORE_TLS_VERSION_1_2	0
 #define KORE_TLS_VERSION_1_0	1
 #define KORE_TLS_VERSION_BOTH	2
 
+#define KORE_RESEED_TIME	(1800 * 1000)
+
 #define errno_s			strerror(errno)
 #define ssl_errno_s		ERR_error_string(ERR_get_error(), NULL)
 
@@ -84,6 +87,7 @@ extern int daemon(int, int);
 #define NETBUF_RECV			0
 #define NETBUF_SEND			1
 #define NETBUF_SEND_PAYLOAD_MAX		8192
+#define SENDFILE_PAYLOAD_MAX		(1024 * 1024 * 10)
 
 #define NETBUF_LAST_CHAIN		0
 #define NETBUF_BEFORE_CHAIN		1
@@ -92,6 +96,7 @@ extern int daemon(int, int);
 #define NETBUF_FORCE_REMOVE	0x02
 #define NETBUF_MUST_RESEND	0x04
 #define NETBUF_IS_STREAM	0x10
+#define NETBUF_IS_FILEREF	0x20
 
 #define X509_GET_CN(c, o, l)					\
 	X509_NAME_get_text_by_NID(X509_get_subject_name(c),	\
@@ -104,6 +109,23 @@ extern int daemon(int, int);
 struct http_request;
 #endif
 
+#define KORE_FILEREF_SOFT_REMOVED	0x1000
+
+struct kore_fileref {
+	int				cnt;
+	int				flags;
+	off_t				size;
+	char				*path;
+	time_t				mtime;
+	u_int64_t			expiration;
+#if !defined(KORE_USE_PLATFORM_SENDFILE)
+	void				*base;
+#else
+	int				fd;
+#endif
+	TAILQ_ENTRY(kore_fileref)	list;
+};
+
 struct netbuf {
 	u_int8_t		*buf;
 	size_t			s_off;
@@ -112,8 +134,13 @@ struct netbuf {
 	u_int8_t		type;
 	u_int8_t		flags;
 
-	void			*owner;
+	struct kore_fileref	*file_ref;
+#if defined(KORE_USE_PLATFORM_SENDFILE)
+	off_t			fd_off;
+	off_t			fd_len;
+#endif
 
+	void			*owner;
 	void			*extra;
 	int			(*cb)(struct netbuf *);
 
@@ -128,7 +155,7 @@ TAILQ_HEAD(netbuf_head, netbuf);
 #define KORE_TYPE_TASK		4
 
 #define CONN_STATE_UNKNOWN		0
-#define CONN_STATE_SSL_SHAKE		1
+#define CONN_STATE_TLS_SHAKE		1
 #define CONN_STATE_ESTABLISHED		2
 #define CONN_STATE_DISCONNECTING	3
 
@@ -143,15 +170,16 @@ TAILQ_HEAD(netbuf_head, netbuf);
 #define CONN_IDLE_TIMER_ACT	0x10
 #define CONN_READ_BLOCK		0x20
 #define CONN_CLOSE_EMPTY	0x40
+#define CONN_WS_CLOSE_SENT	0x80
 
-#define KORE_IDLE_TIMER_MAX	20000
+#define KORE_IDLE_TIMER_MAX	5000
 
 #define WEBSOCKET_OP_CONT	0x00
 #define WEBSOCKET_OP_TEXT	0x01
 #define WEBSOCKET_OP_BINARY	0x02
 #define WEBSOCKET_OP_CLOSE	0x08
 #define WEBSOCKET_OP_PING	0x09
-#define WEBSOCKET_OP_PONG	0x10
+#define WEBSOCKET_OP_PONG	0x0a
 
 #define WEBSOCKET_BROADCAST_LOCAL	1
 #define WEBSOCKET_BROADCAST_GLOBAL	2
@@ -177,8 +205,8 @@ struct connection {
 
 	int			(*handle)(struct connection *);
 	void			(*disconnect)(struct connection *);
-	int			(*read)(struct connection *, int *);
-	int			(*write)(struct connection *, int, int *);
+	int			(*read)(struct connection *, size_t *);
+	int			(*write)(struct connection *, size_t, size_t *);
 
 	u_int8_t		addrtype;
 	union {
@@ -196,7 +224,9 @@ struct connection {
 	struct netbuf		*rnb;
 
 #if !defined(KORE_NO_HTTP)
-	void				*wscbs;
+	struct kore_runtime_call	*ws_connect;
+	struct kore_runtime_call	*ws_message;
+	struct kore_runtime_call	*ws_disconnect;
 	TAILQ_HEAD(, http_request)	http_requests;
 #endif
 
@@ -207,11 +237,37 @@ TAILQ_HEAD(connection_list, connection);
 extern struct connection_list	connections;
 extern struct connection_list	disconnected;
 
+#define KORE_RUNTIME_NATIVE	0
+#define KORE_RUNTIME_PYTHON	1
+
+struct kore_runtime {
+	int	type;
+#if !defined(KORE_NO_HTTP)
+	int	(*http_request)(void *, struct http_request *);
+	int	(*validator)(void *, struct http_request *, const void *);
+	void	(*wsconnect)(void *, struct connection *);
+	void	(*wsdisconnect)(void *, struct connection *);
+	void	(*wsmessage)(void *, struct connection *,
+		    u_int8_t, const void *, size_t);
+#endif
+	void	(*execute)(void *);
+	int	(*onload)(void *, int);
+	void	(*connect)(void *, struct connection *);
+	void	(*configure)(void *, int, char **);
+};
+
+struct kore_runtime_call {
+	void			*addr;
+	struct kore_runtime	*runtime;
+};
+
+extern struct kore_runtime	kore_native_runtime;
+
 struct listener {
-	u_int8_t		type;
-	u_int8_t		addrtype;
-	int			fd;
-	void			(*connect)(struct connection *);
+	u_int8_t			type;
+	u_int8_t			addrtype;
+	int				fd;
+	struct kore_runtime_call	*connect;
 
 	union {
 		struct sockaddr_in	ipv4;
@@ -225,8 +281,11 @@ LIST_HEAD(listener_head, listener);
 
 #if !defined(KORE_NO_HTTP)
 
+#define KORE_PARAMS_QUERY_STRING	0x0001
+
 struct kore_handler_params {
 	char			*name;
+	int			flags;
 	u_int8_t		method;
 	struct kore_validator	*validator;
 
@@ -250,36 +309,53 @@ struct kore_auth {
 #define HANDLER_TYPE_STATIC	1
 #define HANDLER_TYPE_DYNAMIC	2
 
-#endif
+#endif /* !KORE_NO_HTTP */
 
 #define KORE_MODULE_LOAD	1
 #define KORE_MODULE_UNLOAD	2
 
-struct kore_module {
-	void			*handle;
-	char			*path;
-	char			*onload;
-	int			(*ocb)(int);
+#define KORE_MODULE_NATIVE	0
+#define KORE_MODULE_PYTHON	1
 
-	time_t			mtime;
+struct kore_module;
+
+struct kore_module_functions {
+	void			(*free)(struct kore_module *);
+	void			(*reload)(struct kore_module *);
+	int			(*callback)(struct kore_module *, int);
+	void			(*load)(struct kore_module *);
+	void			*(*getsym)(struct kore_module *, const char *);
+};
+
+struct kore_module {
+	void				*handle;
+	char				*path;
+	char				*onload;
+	int				type;
+	time_t				mtime;
+	struct kore_runtime_call	*ocb;
+
+	struct kore_module_functions	*fun;
+	struct kore_runtime		*runtime;
 
 	TAILQ_ENTRY(kore_module)	list;
 };
 
-struct kore_module_handle {
-	char			*path;
-	char			*func;
-	void			*addr;
-	int			type;
-	int			errors;
-	regex_t			rctx;
-	struct kore_domain	*dom;
 #if !defined(KORE_NO_HTTP)
-	struct kore_auth	*auth;
+struct kore_module_handle {
+	char					*path;
+	char					*func;
+	int					type;
+	int					errors;
+	regex_t					rctx;
+	struct kore_domain			*dom;
+	struct kore_runtime_call		*rcall;
+	struct kore_auth			*auth;
+	int					methods;
 	TAILQ_HEAD(, kore_handler_params)	params;
-#endif
 	TAILQ_ENTRY(kore_module_handle)		list;
 };
+#endif
 
 struct kore_worker {
 	u_int8_t			id;
@@ -288,6 +364,8 @@ struct kore_worker {
 	int				pipe[2];
 	struct connection		*msg[2];
 	u_int8_t			has_lock;
+	int				restarted;
+	u_int64_t			time_locked;
 	struct kore_module_handle	*active_hdlr;
 };
 
@@ -300,8 +378,11 @@ struct kore_domain {
 	char					*certfile;
 	char					*certkey;
 	SSL_CTX					*ssl_ctx;
+	int					x509_verify_depth;
 #endif
+#if !defined(KORE_NO_HTTP)
 	TAILQ_HEAD(, kore_module_handle)	handlers;
+#endif
 	TAILQ_ENTRY(kore_domain)		list;
 };
 
@@ -313,15 +394,15 @@ TAILQ_HEAD(kore_domain_h, kore_domain);
 #define KORE_VALIDATOR_TYPE_FUNCTION	2
 
 struct kore_validator {
-	u_int8_t		type;
-	char			*name;
-	char			*arg;
-	regex_t			rctx;
-	int			(*func)(struct http_request *, char *);
+	u_int8_t			type;
+	char				*name;
+	char				*arg;
+	regex_t				rctx;
+	struct kore_runtime_call	*rcall;
 
 	TAILQ_ENTRY(kore_validator)	list;
 };
-#endif
+#endif /* !KORE_NO_HTTP */
 
 #define KORE_BUF_OWNER_API	0x0001
 
@@ -356,13 +437,6 @@ struct kore_pool {
 	LIST_HEAD(, kore_pool_entry)	freelist;
 };
 
-struct kore_wscbs {
-	void		(*connect)(struct connection *);
-	void		(*message)(struct connection *, u_int8_t,
-			    void *, size_t);
-	void		(*disconnect)(struct connection *);
-};
-
 struct kore_timer {
 	u_int64_t	nextrun;
 	u_int64_t	interval;
@@ -376,10 +450,15 @@ struct kore_timer {
 #define KORE_WORKER_KEYMGR	0
 
 /* Reserved message ids, registered on workers. */
-#define KORE_MSG_ACCESSLOG	1
-#define KORE_MSG_WEBSOCKET	2
-#define KORE_MSG_KEYMGR_REQ	3
-#define KORE_MSG_KEYMGR_RESP	4
+#define KORE_MSG_ACCESSLOG		1
+#define KORE_MSG_WEBSOCKET		2
+#define KORE_MSG_KEYMGR_REQ		3
+#define KORE_MSG_KEYMGR_RESP		4
+#define KORE_MSG_SHUTDOWN		5
+#define KORE_MSG_ENTROPY_REQ		6
+#define KORE_MSG_ENTROPY_RESP		7
+#define KORE_MSG_CERTIFICATE		8
+#define KORE_MSG_CERTIFICATE_REQ	9
 
 /* Predefined message targets. */
 #define KORE_MSG_PARENT		1000
@@ -389,17 +468,24 @@ struct kore_msg {
 	u_int8_t	id;
 	u_int16_t	src;
 	u_int16_t	dst;
-	u_int32_t	length;
+	size_t		length;
 };
 
 #if !defined(KORE_NO_TLS)
 struct kore_keyreq {
 	int		padding;
 	char		domain[KORE_DOMAINNAME_LEN];
-	u_int8_t	domain_len;
+	u_int16_t	domain_len;
 	u_int16_t	data_len;
 	u_int8_t	data[];
 };
+
+struct kore_x509_msg {
+	char		domain[KORE_DOMAINNAME_LEN];
+	u_int16_t	domain_len;
+	size_t		data_len;
+	u_int8_t	data[];
+};
 #endif
 
 #if !defined(KORE_SINGLE_BINARY)
@@ -410,20 +496,24 @@ extern pid_t	kore_pid;
 extern int	foreground;
 extern int	kore_debug;
 extern int	skip_chroot;
-extern char	*chroot_path;
 extern int	skip_runas;
-extern char	*runas_user;
 extern char	*kore_pidfile;
+extern char	*kore_root_path;
+extern char	*kore_runas_user;
 extern char	*kore_tls_cipher_list;
-extern int	tls_version;
 
 #if !defined(KORE_NO_TLS)
+extern int	tls_version;
 extern DH	*tls_dhparam;
+extern char	*rand_file;
+extern char	*keymgr_runas_user;
+extern char	*keymgr_root_path;
 #endif
 
 extern u_int8_t			nlisteners;
 extern u_int16_t		cpu_count;
 extern u_int8_t			worker_count;
+extern const char		*kore_version;
 extern u_int8_t			worker_set_affinity;
 extern u_int32_t		worker_rlimit_nofiles;
 extern u_int32_t		worker_max_connections;
@@ -439,17 +529,16 @@ extern struct kore_domain_h	domains;
 extern struct kore_domain	*primary_dom;
 extern struct kore_pool		nb_pool;
 
-void		kore_cli_usage(int);
-int		kore_cli_main(int, char **);
-
 void		kore_signal(int);
+void		kore_signal_setup(void);
 void		kore_worker_wait(int);
 void		kore_worker_init(void);
+void		kore_worker_make_busy(void);
 void		kore_worker_shutdown(void);
-void		kore_worker_privdrop(void);
 void		kore_worker_dispatch_signal(int);
 void		kore_worker_spawn(u_int16_t, u_int16_t);
 void		kore_worker_entry(struct kore_worker *);
+void		kore_worker_privdrop(const char *, const char *);
 
 struct kore_worker	*kore_worker_data(u_int8_t);
 
@@ -467,6 +556,10 @@ void		kore_platform_schedule_write(int, void *);
 void		kore_platform_event_schedule(int, int, int, void *);
 void		kore_platform_worker_setcpu(struct kore_worker *);
 
+#if defined(KORE_USE_PLATFORM_SENDFILE)
+int		kore_platform_sendfile(struct connection *, struct netbuf *);
+#endif
+
 void		kore_accesslog_init(void);
 void		kore_accesslog_worker_init(void);
 int		kore_accesslog_write(const void *, u_int32_t);
@@ -484,6 +577,7 @@ void		kore_timer_remove(struct kore_timer *);
 struct kore_timer	*kore_timer_add(void (*cb)(void *, u_int64_t),
 			    u_int64_t, void *, int);
 
+int		kore_sockopt(int, int, int);
 void		kore_listener_cleanup(void);
 int		kore_server_bind(const char *, const char *, const char *);
 #if !defined(KORE_NO_TLS)
@@ -495,8 +589,8 @@ void			kore_connection_init(void);
 void			kore_connection_cleanup(void);
 void			kore_connection_prune(int);
 struct connection	*kore_connection_new(void *);
-void			kore_connection_check_timeout(void);
 int			kore_connection_nonblock(int, int);
+void			kore_connection_check_timeout(u_int64_t);
 int			kore_connection_handle(struct connection *);
 void			kore_connection_remove(struct connection *);
 void			kore_connection_disconnect(struct connection *);
@@ -512,10 +606,16 @@ void		kore_log_init(void);
 
 void		*kore_malloc(size_t);
 void		kore_parse_config(void);
+void		kore_parse_config_file(FILE *);
 void		*kore_calloc(size_t, size_t);
 void		*kore_realloc(void *, size_t);
 void		kore_free(void *);
 void		kore_mem_init(void);
+void		kore_mem_cleanup(void);
+void		kore_mem_untag(void *);
+void		*kore_mem_lookup(u_int32_t);
+void		kore_mem_tag(void *, u_int32_t);
+void		*kore_malloc_tagged(size_t, u_int32_t);
 
 void		*kore_pool_get(struct kore_pool *);
 void		kore_pool_put(struct kore_pool *, void *);
@@ -523,26 +623,29 @@ void		kore_pool_init(struct kore_pool *, const char *,
 		    size_t, size_t);
 void		kore_pool_cleanup(struct kore_pool *);
 
-time_t		kore_date_to_time(char *);
 char		*kore_time_to_date(time_t);
 char		*kore_strdup(const char *);
-void		kore_log(int, const char *, ...);
+time_t		kore_date_to_time(const char *);
+void		kore_log(int, const char *, ...)
+		    __attribute__((format (printf, 2, 3)));
 u_int64_t	kore_strtonum64(const char *, int, int *);
 size_t		kore_strlcpy(char *, const char *, const size_t);
 void		kore_server_disconnect(struct connection *);
-int		kore_split_string(char *, char *, char **, size_t);
+int		kore_split_string(char *, const char *, char **, size_t);
 void		kore_strip_chars(char *, const char, char **);
 int		kore_snprintf(char *, size_t, int *, const char *, ...);
 long long	kore_strtonum(const char *, int, long long, long long, int *);
-int		kore_base64_encode(u_int8_t *, size_t, char **);
-int		kore_base64_decode(char *, u_int8_t **, size_t *);
+double		kore_strtodouble(const char *, long double, long double, int *);
+int		kore_base64_encode(const void *, size_t, char **);
+int		kore_base64_decode(const char *, u_int8_t **, size_t *);
 void		*kore_mem_find(void *, size_t, void *, size_t);
 char		*kore_text_trim(char *, size_t);
 char		*kore_read_line(FILE *, char *, size_t);
 
 #if !defined(KORE_NO_HTTP)
 void		kore_websocket_handshake(struct http_request *,
-		    struct kore_wscbs *);
+		    const char *, const char *, const char *);
+int		kore_websocket_send_clean(struct netbuf *);
 void		kore_websocket_send(struct connection *,
 		    u_int8_t, const void *, size_t);
 void		kore_websocket_broadcast(struct connection *,
@@ -554,10 +657,24 @@ void		kore_msg_worker_init(void);
 void		kore_msg_parent_init(void);
 void		kore_msg_parent_add(struct kore_worker *);
 void		kore_msg_parent_remove(struct kore_worker *);
-void		kore_msg_send(u_int16_t, u_int8_t, const void *, u_int32_t);
+void		kore_msg_send(u_int16_t, u_int8_t, const void *, size_t);
 int		kore_msg_register(u_int8_t,
 		    void (*cb)(struct kore_msg *, const void *));
 
+#if !defined(KORE_NO_HTTP)
+void		kore_filemap_init(void);
+void		kore_filemap_resolve_paths(void);
+int		kore_filemap_create(struct kore_domain *, const char *,
+		    const char *);
+extern char	*kore_filemap_ext;
+extern char	*kore_filemap_index;
+#endif
+
+void			kore_fileref_init(void);
+struct kore_fileref	*kore_fileref_get(const char *);
+struct kore_fileref	*kore_fileref_create(const char *, int, off_t, time_t);
+void			kore_fileref_release(struct kore_fileref *);
+
 void		kore_domain_init(void);
 void		kore_domain_cleanup(void);
 int		kore_domain_new(char *);
@@ -568,19 +685,39 @@ void		kore_module_reload(int);
 void		kore_module_onload(void);
 int		kore_module_loaded(void);
 void		kore_domain_closelogs(void);
-void		*kore_module_getsym(const char *);
+void		*kore_module_getsym(const char *, struct kore_runtime **);
 void		kore_domain_load_crl(void);
 void		kore_domain_keymgr_init(void);
-void		kore_module_load(const char *, const char *);
-void		kore_domain_sslstart(struct kore_domain *);
+void		kore_module_load(const char *, const char *, int);
 void		kore_domain_callback(void (*cb)(struct kore_domain *));
+void		kore_domain_tlsinit(struct kore_domain *, const void *, size_t);
+#if !defined(KORE_NO_HTTP)
 int		kore_module_handler_new(const char *, const char *,
 		    const char *, const char *, int);
 void		kore_module_handler_free(struct kore_module_handle *);
-
-struct kore_domain		*kore_domain_lookup(const char *);
 struct kore_module_handle	*kore_module_handler_find(const char *,
 				    const char *);
+#endif
+
+struct kore_runtime_call	*kore_runtime_getcall(const char *);
+
+void	kore_runtime_execute(struct kore_runtime_call *);
+int	kore_runtime_onload(struct kore_runtime_call *, int);
+void	kore_runtime_configure(struct kore_runtime_call *, int, char **);
+void	kore_runtime_connect(struct kore_runtime_call *, struct connection *);
+#if !defined(KORE_NO_HTTP)
+int	kore_runtime_http_request(struct kore_runtime_call *,
+	    struct http_request *);
+int	kore_runtime_validator(struct kore_runtime_call *,
+	    struct http_request *, const void *);
+void	kore_runtime_wsconnect(struct kore_runtime_call *, struct connection *);
+void	kore_runtime_wsdisconnect(struct kore_runtime_call *,
+	    struct connection *);
+void	kore_runtime_wsmessage(struct kore_runtime_call *,
+	    struct connection *, u_int8_t, const void *, size_t);
+#endif
+
+struct kore_domain		*kore_domain_lookup(const char *);
 
 #if !defined(KORE_NO_HTTP)
 void		kore_validator_init(void);
@@ -588,7 +725,7 @@ void		kore_validator_reload(void);
 int		kore_validator_add(const char *, u_int8_t, const char *);
 int		kore_validator_run(struct http_request *, const char *, char *);
 int		kore_validator_check(struct http_request *,
-		    struct kore_validator *, void *);
+		    struct kore_validator *, const void *);
 struct kore_validator	*kore_validator_lookup(const char *);
 #endif
 
@@ -604,13 +741,14 @@ void		net_write64(u_int8_t *, u_int64_t);
 
 void		net_init(void);
 void		net_cleanup(void);
+struct netbuf	*net_netbuf_get(void);
 int		net_send(struct connection *);
 int		net_send_flush(struct connection *);
 int		net_recv_flush(struct connection *);
-int		net_read(struct connection *, int *);
-int		net_read_ssl(struct connection *, int *);
-int		net_write(struct connection *, int, int *);
-int		net_write_ssl(struct connection *, int, int *);
+int		net_read(struct connection *, size_t *);
+int		net_read_tls(struct connection *, size_t *);
+int		net_write(struct connection *, size_t, size_t *);
+int		net_write_tls(struct connection *, size_t, size_t *);
 void		net_recv_reset(struct connection *, size_t,
 		    int (*cb)(struct netbuf *));
 void		net_remove_netbuf(struct netbuf_head *, struct netbuf *);
@@ -621,6 +759,7 @@ void		net_recv_expand(struct connection *c, size_t,
 void		net_send_queue(struct connection *, const void *, size_t);
 void		net_send_stream(struct connection *, void *,
 		    size_t, int (*cb)(struct netbuf *), struct netbuf **);
+void		net_send_fileref(struct connection *, struct kore_fileref *);
 
 void		kore_buf_free(struct kore_buf *);
 struct kore_buf	*kore_buf_alloc(size_t);
@@ -636,11 +775,10 @@ void	kore_buf_appendv(struct kore_buf *, const char *, va_list);
 void	kore_buf_replace_string(struct kore_buf *, char *, void *, size_t);
 
 void	kore_keymgr_run(void);
-void	kore_keymgr_cleanup(void);
+void	kore_keymgr_cleanup(int);
 
-#if defined(KORE_SINGLE_BINARY)
-void	kore_onload(void);
-#endif
+void	kore_worker_configure(void);
+void	kore_parent_configure(int, char **);
 
 #if defined(__cplusplus)
 }
